<?php
class RESTDispatcher {
	/**
	 * Current path override
	 * @var string
	 */
	private static $root_path='modules';
	/**
	 * Al previous path overrides, to allow revert
	 * @var array
	 */
	private static $prev_path=array();
	/**
	 * Current namespace override
	 * @var string
	 */
	private static $root_namespace='';
	/**
	 * Al previous namespace overrides, to allow revert
	 * @var array
	 */
	private static $prev_namespaces=array();
	/**
	 * Current URI override
	 * @var string
	 */
	private static $root_uri='';
	/**
	 * Al previous URI overrides, to allow revert
	 * @var array
	 */
	private static $prev_uri=array();
	/**
	 * The file to load
	 * @var string
	 */
	private $file=NULL;
	/**
	 * Something was already dispatched, stop checking
	 * @var bool
	 */
	private $kill=false;
	/**
	 * Assigns a path to a dispatcher file
	 * @param string $uri Path structure
	 * @param string $file File to load
	 * @throws NotImplementedException
	 * @throws MethodNotAllowedException
	 * @return RESTDispatcher
	 */
	function __construct($uri,$file) {
		if($uri===NULL&&$file===NULL||$this->kill) return $this;
		if($file!='') $ns=self::$root_namespace.'.';
		else $ns=self::$root_namespace;
		$input=REST::input();
		$current_uri=$input->getPart();
		$uri=self::$root_uri.$uri;
		if(substr($uri, -1)=='/') $uri=substr($uri,0,-1); //Remove last / from uri
		$uri=preg_quote($uri,'/');
		$search=array('\\{GUID\\}',												'\\{ANID\\}',			'\\{NID\\}',	'\\{ALPHA\\}', '\\{\\}');
		$replace=array('[0-9A-Fa-f]{8}-([0-9A-Fa-f]{4}-){3}[0-9A-Fa-f]{12}',	'[0-9A-Za-z]+',			'[0-9]+',		'[A-Za-z]+',	'[^\\/]+');
		$pattern=str_replace($search,$replace,$uri).'\\/?';
		if(preg_match('/'.$pattern.'/', $current_uri))  {
			$this->file=self::$root_path.'/'.$ns.$file.'.php';
			if(!file_exists($this->file)) throw new NotImplementedException('Module '.$ns.$file.' is not available');
			$return=$this->launch();
			if($return===false) throw new MethodNotAllowedException('Invalid method for "'.$ns.$file.'"');
			$this->kill=true;
		}
		return $this;
	}
	/**
	 * Open the file to be executed, assign right parameters
	 * @return bool The return value of the file, true means OK, false MethodNotAllowedException
	 */
	private function launch() {
		global $local_connection,$remote_connection;
		return require $this->file;
	}
	/**
	 * Set a new namespace.
	 * @param string $new The new namespace. If a dot (.) is prepended, the previous namespace is extended
	 * @return RESTDispatcher
	 */
	function override_namespace($new) {
		if($this->kill) return $this;
		self::$prev_namespaces[]=self::$root_namespace;
		if(strpos($new, '.')===0) $new=self::$root_namespace.$new;
		self::$root_namespace=$new;
		return $this;
	}
	/**
	 * Set a new path to look for files in
	 * @param string $new The new path. If './' is prepended, the previous path is extended
	 * @throws RuntimeException When path is not a directory
	 * @return RESTDispatcher
	 */
	function override_path($new) {
		if($this->kill) return $this;
		self::$prev_path[]=self::$root_path;
		if(strpos($new, './')===0) $new=self::$root_path.'/'.$new;
		if(!is_dir($new)) throw new RuntimeException('The path '.$new.' does not exist');
		self::$root_path=$new;
		return $this;
	}
	/**
	 * Set a new uri to put before the give part
	 * @param string $new The new URI. If './' is prepended, the previous URI is extended
	 * @return RESTDispatcher
	 */
	function override_uri($new) {
		if($this->kill) return $this;
		self::$prev_uri[]=self::$root_uri;
		if(strpos($new,'./')===0) $new=self::$root_uri.substr($new, 2);
		self::$root_uri=$new.'/';
		return $this;
	}
	/**
	 * Overrides namespace, uri and path
	 * @param string $namespace The new namespace
	 * @param string $uri The new URI
	 * @param string $path The new path
	 * @see RESTDispatcher::override_namespace()
	 * @see RESTDispatcher::override_path()
	 * @see RESTDispatcher::override_uri()
	 * @uses override_namespace()
	 * @uses override_path()
	 * @uses override_uri()
	 * @return RESTDispatcher
	 */
	function override($namespace=NULL,$uri=NULL,$path=NULL) {
		if($this->kill) return $this;
		if(!is_null($namespace)) $this->override_namespace($namespace);
		if(!is_null($uri)) $this->override_uri($uri);
		if(!is_null($path)) $this->override_path($path);
		return $this;
	}
	/**
	 * Puts the previous namespace back
	 * @return RESTDispatcher
	 */
	function revert_namespace() {
		if($this->kill) return $this;
		self::$root_namespace=array_pop(self::$prev_namespaces);
		return $this;
	}
	/**
	 * Puts the previous path back
	 * @return RESTDispatcher
	 */
	function revert_path() {
		if($this->kill) return $this;
		self::$root_path=array_pop(self::$prev_path);
		return $this;
	}
	/**
	 * Puts the previous URI back
	 * @return RESTDispatcher
	 */
	function revert_uri() {
		if($this->kill) return $this;
		self::$root_uri=array_pop(self::$prev_uri);
		return $this;
	}
	/**
	 * Assigns a path to a dispatcher file
	 * @uses RESTDispatcher::__construct();
	 * @param string $uri Path structure
	 * @param string $file File to load
	 * @return RESTDispatcher
	 */
	function dispatch($uri,$file) {
		if($this->kill) return $this;
		return $this->__construct($uri,$file);
	}

	/**
	 * Debug function
	 * @internal
	 */
	function debug() {
		echo 'Path:'.self::$root_path."\r\n";
		var_dump(self::$prev_path);
		echo 'NS:'.self::$root_namespace."\r\n";
		var_dump(self::$prev_namespaces);
		echo 'URI:'.self::$root_uri."\r\n";
		var_dump(self::$prev_uri);
		return $this;
	}
}