<?php
class ftpfs extends fs {
	const version='1.1';
	/**
	 * Holds the FTP-connection
	 * @var resource
	 */
	protected $ftplink;
	/**
	 * Holds the basedirectory for file storage
	 * @var string
	 */
	protected $base_dir;
	/**
	 * Constructor for the ftpfs class
	 * @param string $user The FTP username
	 * @param string $pass The FTP password
	 * @param string $host The FTP host
	 * @param string $base_dir The root directory for FTPfs
	 * @param int $port The port of the FTP-server
	 */
	function __construct($user='anonymous',$pass='',$host='localhost',$base_dir='/',$port=21) {
		$this->ftplink=ftp_connect($host,$port);
		ftp_login($this->ftplink, $user, $pass);
		$this->base_dir=$base_dir.'/';
		if(!@ftp_chdir($this->ftplink, $base_dir)) {
			$this->trigger_error('FTPfs: Base directory does not exist',E_USER_WARNING);
		}
	}
	/**
	 * @internal does nothing
	 */
	final function setuser($user=NULL,$group=0,$groups=array()) {
	}
	/**
	 * @internal FTP does not support this
	 */
	final function chgrp($path,$group) {
		return false;
	}
	final function chmod($path,$mode) {
		$mode = octdec( str_pad($mode,4,'0',STR_PAD_LEFT) );
		$mode=(int)$mode;
		return ftp_chmod($this->ftplink, $mode, $this->base_dir.$filename);
	}
	/**
	 * @internal FTP does not support this
	 */
	final function chown($path,$user) {
		return false;
	}
	final function copy($source,$dest) {
		if(($source=$this->simplify_path($source))===NULL)  { $this->trigger_error("FTPfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(($dest=$this->simplify_path($dest))===NULL)  { $this->trigger_error("FTPfs: ".func_get_arg(1)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->is_readable($source)) { $this->trigger_error("FTPfs: $source is not readable",E_USER_WARNING); return false; }
		$ddir=$this->dirname($dest);
		if(!$this->is_writable($ddir)) { $this->trigger_error("FTPfs: $ddir is not writable",E_USER_WARNING); return false; }
		if($this->file_exists($dest)) { $this->trigger_error("FTPfs: $dest already exists",E_USER_WARNING); return false; }
		return $this->tmpfile_ftp_put($this->tmpfile_ftp_get($source),$dest);
	}
	final function file($path,$flags=0) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("FTPfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->is_readable($path)) { $this->trigger_error("FTPfs: $path is not readable",E_USER_WARNING); return false; }
		return file($this->tmpfile_ftp_get($path,$flags));
	}
	final function file_exists($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("FTPfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		return ($this->is_file($path)||$this->is_dir($path));
	}
	final function file_get_contents($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("FTPfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->is_readable($path)) { $this->trigger_error("FTPfs: $path is not readable",E_USER_WARNING); return false; }
		return ($this->tmpfile_read($this->tmpfile_ftp_put($this->tmpfile_write(''),$path)));
	}
	final function file_put_contents($path,$data) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("FTPfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		$dir=$this->dirname($path);
		if(!$this->is_writable($dir)) { $this->trigger_error("FTPfs: $dir is not writable",E_USER_WARNING); return false; }
		return ($this->tmpfile_ftp_put($this->tmpfile_write($data),$path));
	}
	final function filegroup($path) {
		return $this->getinfo($path,'group');
	}
	final function filemtime($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("FTPfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		return (ftp_mdtm($this->ftplink,$this->base_dir.$path));
	}
	final function fileowner($path) {
		return $this->getinfo($path,'owner');
	}
	final function fileperms($path) {
		return $this->getinfo($path,'rights');
	}
	final function filesize($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("FTPfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->file_exists($path)) { $this->trigger_error("FTPfs: $path does not exist",E_USER_WARNING); return false; }
		return (ftp_size($this->ftplink,$this->base_dir.$path));
	}
	final function filetype($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("FTPfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->file_exists($path)) { $this->trigger_error("FTPfs: $path does not exist",E_USER_WARNING); return false; }
		if($this->is_dir($path)) return 'directory';
		if($this->is_file($path)) return 'file';
	}
	final function is_dir($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("FTPfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if($path==''||$path=='/'||@ftp_chdir($this->ftplink,$this->base_dir.$path)) $dir=true;
		else $dir=false;
		@ftp_chdir($this->ftplink,$this->base_dir);
		return $dir;
	}
	final function is_empty($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("FTPfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->is_dir($path)) return true; //File doesn't contain files
		return (count(ftp_nlist($this->ftplink,$this->base_dir.$path))==0);
	}
	final function is_executable($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("FTPfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->file_exists($path)) return false;
		return true;
	}
	final function is_file($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("FTPfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		return(ftp_size($this->ftplink,$this->base_dir.$path)!==-1);
	}
	final function is_readable($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("FTPfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->file_exists($path)) return false;
		return true;
	}
	final function is_writable($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("FTPfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->file_exists($path)) return false;
		return true;
	}
	final function mkdir($path,$mode=NULL) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("FTPfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		return (ftp_mkdir($this->ftplink,$this->base_dir.$path)!==false);
	}
	final function mkfile($path,$mode=NULL,$mime=NULL) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("FTPfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		return $this->tmpfile_ftp_put($this->tmpfile_write(''),$path);
	}
	final function move_uploaded_file($source,$dest) {
		if(!is_uploaded_file($source)) { $this->trigger_error("FTPfs: $source is not a valid upload file",E_USER_WARNING); return false; }
		if(($dest=$this->simplify_path($dest))===NULL)  { $this->trigger_error("FTPfs: ".func_get_arg(1)." is an invalid pathname",E_USER_WARNING); return false; }
		return $this->tmpfile_ftp_put($source,$dest);
	}
	final function readdir_full($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("FTPfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->is_dir($path)) { $this->trigger_error("FTPfs: $path is not a directory",E_USER_WARNING); return NULL; }
		if(!$this->is_readable($path)) { $this->trigger_error("FTPfs: $path is not readable",E_USER_WARNING); return NULL; }
		$out=array();
		$dirlist=$this->parse_ftp_rawlist(ftp_rawlist($this->ftplink, $this->base_dir.$path));
		$out[]=$this->getinfo($path.'/..');
		$out[0]['filename']='..';
		$out[]=$this->getinfo($path);
		$out[1]['filename']='.';
		foreach($dirlist as $finfo) {
			$info['type']=$finfo['type'];
			$info['filename']=$finfo['filename'];
			$info['time']=strtotime($finfo['time']);
			$info['rights']=$finfo['rights'];
			$info['mime']=$this->mime_content_type($path);
			$info['size']=$info['type']=='d'?0:$finfo['size'];
			$info['owner']=$finfo['owner'];
			$info['group']=$finfo['group'];
			if($info['owner']===NULL) $info['owner']='nobody';
			if($info['group']===NULL) $info['group']='nobody';
			if($info['type']=='d') unset($info['mime']); //Dir's have no mimetype
			$out[]=$info;
		}
		return $out;
	}
	final function readfile($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("FTPfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		$c=$this->file_get_contents($path);
		echo $c;
		return strlen($c);
	}
	final function rename($oldpath,$newpath) {
		if(($oldpath=$this->simplify_path($oldpath))===NULL)  { $this->trigger_error("FTPfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(($newpath=$this->simplify_path($newpath))===NULL)  { $this->trigger_error("FTPfs: ".func_get_arg(1)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->is_readable($oldpath)) { $this->trigger_error("FTPfs: $oldpath is not readable",E_USER_WARNING); return false; }
		if(!$this->is_writable($oldpath)) { $this->trigger_error("FTPfs: $olpath is not writable", E_USER_WARNING); return false;	}
		$oldfilename=$this->basename($oldpath);
		$newfilename=$this->basename($newpath);
		$olddir=$this->dirname($oldpath);
		$newdir=$this->dirname($newpath);
		if(!$this->is_dir($newdir)) $newdir=$this->simplify_path($olddir.'/'.$newdir); //Relative?
		if(!$this->is_writable($newdir)) { $this->trigger_error("FTPfs: $newdir is not writable",E_USER_WARNING); return false; }
		if($this->file_exists($newpath)) { $this->trigger_error("FTPfs: $newpath already exists",E_USER_WARNING); return false; }
		if(preg_match("|^$oldpath|",$newpath)) { $this->trigger_error("FTPfs: $newpath is a subdirectory of $oldpath",E_USER_WARNING); return false; }
		return ftp_rename($this->ftplink,$this->base_dir.$oldpath,$this->base_dir.$newpath);
	}
	final function rmdir($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->is_writable($path)) { $this->trigger_error("MySQLfs: $path is not writable",E_USER_WARNING); return false; }
		if(!$this->is_dir($path)) { $this->trigger_error("MySQLfs: '$path' is not a directory",E_USER_WARNING); return false; }
		if(!$this->is_empty($path)) { $this->trigger_error("MySQLfs: '$path' is not empty",E_USER_WARNING); return false; }
		return ftp_rmdir($this->ftplink,$this->base_dir.$path);
	}
	final function scandir($path,$sorting_order=0) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->file_exists($path)) { $this->trigger_error("MySQLfs: $path does not exist",E_USER_WARNING); return false; }
		if(!$this->is_dir($path)) { $this->trigger_error("MySQLfs: $path is not a directory",E_USER_WARNING); return false; }
		if(!$this->is_readable($path)) { $this->trigger_error("MySQLfs: $path is not readable",E_USER_WARNING); return false; }
		$files=ftp_nlist($this->ftplink,$this->base_dir.$path);
		if($sorting_order==1) $files=array_reverse($files);
		return $files;
	}
	final function unlink($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("FTPfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->is_writable($path)) { $this->trigger_error("FTPfs: $path is not writable",E_USER_WARNING); return false; }
		if(!$this->is_file($path)) { $this->trigger_error("FTPfs: '$path' is not a file",E_USER_WARNING); return false; }
		return ftp_delete($this->ftplink,$this->base_dir.$path);
	}
	final function sizetohuman($size) {
		if($size<0) $size=0;
		return parent::sizetohuman($size);
	}
	/**
	 * Gets information about a path
	 * @param string $path The path where to retrieve the information from
	 * @param string $what Which field to return
	 * @return mixed array when $what is not defined, string when $what is defined
	 */
	final function getinfo($path,$what=NULL) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("FTPfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		$exists=$this->file_exists($path);
		$path=$this->base_dir.$path;
		$file=$this->basename($path);
		$dirlist=$this->parse_ftp_rawlist(ftp_rawlist($this->ftplink, $path.'/..'));
		foreach($dirlist as $finfo) {
			if($finfo['filename']==$file) break;
		}
		$info['type']=$finfo['type'];
		$info['filename']=$finfo['filename'];
		$info['time']=strtotime($finfo['time']);
		$info['rights']=$finfo['rights'];
		$info['mime']=$this->mime_content_type($path);
		$info['size']=$info['type']=='d'?0:$finfo['size'];
		$info['owner']=$finfo['owner'];
		$info['group']=$finfo['group'];
		if($info['owner']===NULL) $info['owner']='nobody';
		if($info['group']===NULL) $info['group']='nobody';
		if($info['type']=='d') unset($info['mime']); //Dir's have no mimetype
		if($what!==NULL) $info=$info[$what];
		return $info;
	}
	//
	final private function tmpfile_ftp_get($ftp_path) {
		$tmpfile=tempnam(__DIR__,'ftpfs');
		if(!ftp_get($this->ftplink,$tmpfile,$this->base_dir.$ftp_path,FTP_BINARY)) return false;
		return $tmpfile;
	}
	final private function tmpfile_ftp_put($file,$ftp_path) {
		if(!ftp_put($this->ftplink,$this->base_dir.$ftp_path,$file,FTP_BINARY)) {return false;}
		unlink($file);
		return true;
	}
	final private function tmpfile_write($data) {
		$tmpfile=tempnam(__DIR__,'ftpfs');
		file_put_contents($tmpfile,$data);
		return $tmpfile;
	}
	final private function tmpfile_read($file) {
		$data=file_get_contents($file);
		unlink($file);
		return $data;
	}
	/**
	 * Parse the FTP rawlist
	 * @param array $List returned array by ftp_rawlist
	 * @param bool $Win Is it windows?
	 */
	final private function parse_ftp_rawlist($List,$Win=false) {
		// :: JaZz phpFramework
		// :: V1.1
		// :: 13/03/2009 15:35:22
		$Output = array();
		$i = 0;
		if ($Win) {
			foreach ($List as $Current) {
				ereg('([0-9]{2})-([0-9]{2})-([0-9]{2}) +([0-9]{2}):([0-9]{2})(AM|PM) +([0-9]+|) +(.+)', $Current, $Split);
				if (is_array($Split)) {
					if ($Split[3] < 70) {
						$Split[3] += 2000;
					}
					else {
						$Split[3] += 1900;
					}
					$Output[$i]['type']	 = ($Split[7] == ''?'d':'f');
					$Output[$i]['size']	  = $Split[7];
					$Output[$i]['month']	 = $Split[1];
					$Output[$i]['day']	   = $Split[2];
					$Output[$i]['time/year'] = $Split[3];
					$Output[$i]['filename']	  = $Split[8];
					$i++;
				}
			}
			return !empty($Output) ? $Output : false;
		}
		else {
			unset($List[0]);
			foreach ($List as $Current) {
				$Split = preg_split('[ ]', $Current, 9, PREG_SPLIT_NO_EMPTY);
					$Output[$i]['type']	 = ($Split[0] {0} === 'd'?'d':'f');
					$Output[$i]['rights']	 = $this->perm_to_oct($Split[0]);
					$Output[$i]['owner']	 = $Split[2];
					$Output[$i]['group']	 = $Split[3];
					$Output[$i]['size']	  = $Split[4];
					$Output[$i]['time']	 = $Split[5].' '.$Split[6].' '.$Split[7];
					$Output[$i]['filename'] = $Split[8];
					$i++;
			}
			return !empty($Output) ? $Output : false;
		}
	}
	final private function perm_to_oct($perm) {
		$perm=substr($perm, 1);
		$blocks=str_split($perm,3);
		$permsa=array();
		for ($i = 0; $i < 3; $i++) {
			$block=$blocks[$i];
			$bit=0;
			if($block[0]=='r') $bit=$bit|4;
			if($block[1]=='w') $bit=$bit|2;
			if($block[2]=='x') $bit=$bit|1;
			$permsa[$i]=$bit;
		}
		return implode('', $permsa);
	}
}
