<?php
/**
 * This is the interface to HybridFS
 * @author Lars Vierbergen
 * @package HybridFS
 */
class HybridFS extends fs {
	/**
	 * The main group of the user
	 * @var string
	 */
	public $group=0;
	/**
	 * The userid if the users
	 * @var string
	 */
	public $user=NULL;
	/**
	 * All groups the user is a member of
	 * @var array
	 */
	public $groups=array();
	/**
	 * Holds the MySQL connection
	 * @var mysqli
	 */
	protected $connection;
	/**
	 * Holds the table for HybridFS
	 * @var string
	 */
	protected $table;
	/**
	 * Holds the object interface to the actual filesystem
	 * @var fs
	 */
	protected $realfs=NULL;
	/**
	 * Startup function for the class. Registers MySQL connection and table settings
	 * @param MySQLi $conn The MySQLi link to the database
	 * @param fs $realfs The object interface to the data storage filesystem
	 * @param string $tbl The name of the index table, the datatable will have 'data' appended
	 */
	function __construct($conn,$realfs,$tbl='index') {
		$this->connection=$conn;
		$this->realfs=$realfs;
		$this->table=$tbl;
		/**
		 * @internal Avoid warnings about strtotime calls (bug #815161)
		 * @ignore
		 */
		date_default_timezone_set(@date_default_timezone_get());
	}
	/**
	 * Sets $user and $group (user's main group) and $groups (all groups user is member of)
	 * @param string $user The identifier of the current user
	 * @param string $group The identifier of the main group of the user
	 * @param array $groups All groups the user is member of
	 */
	final function setuser($user=NULL,$group=0,$groups=array()) {
		$this->user=$user;
		$this->group=$group;
		$this->groups=$groups;
	}
	/**
	 * Reads the directory into an array
	 * @param string $path The path to read
	 * @return array Array with nested arrays containing information about the file
	 */
	final function readdir_full($path){
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("HybridFS: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->is_dir($path)) { $this->trigger_error("HybridFS: $path is not a directory",E_USER_WARNING); return NULL; }
		if(!$this->is_readable($path)) { $this->trigger_error("HybridFS: $path is not readable",E_USER_WARNING); return NULL; }
		$result=$this->connection->query("SELECT * FROM `{$this->table}` WHERE `path` = '$path' AND `filename`!=''");
		$out=array();
		$out[]=$this->getinfo($path.'/..'); $out[0]['filename']='..';
		$out[]=$this->getinfo($path); $out[1]['filename']='.';
		while($row=$result->fetch_assoc()) {
			if(!$this->is_readable($path.'/'.$row['filename'])) continue;
			$out[]=$row;
		}
		$result->free();
		return $out;
	}
	/**
	 * Gets information about a path
	 * @param string $path The path where to retrieve the information from
	 * @param string $what Which field to return
	 * @return mixed array when $what is not defined, string when $what is defined
	 */
	final function getinfo($path,$what=NULL) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("HybridFS: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		$filename=$this->basename($path);
		$dir=$this->dirname($path);
		if($dir=="\\"||$dir=='.') $dir='';
		$result=$this->connection->query("SELECT * FROM `{$this->table}` WHERE `path`='$dir' AND `filename` = '$filename'");
		$info=$result->fetch_assoc();
		$result->free();
		if($info['owner']===NULL) $info['owner']='nobody';
		if($info['group']===NULL) $info['group']='nobody';
		if($info['type']=="d") unset($info['mime']); //Dir's have no mimetype
		if($what!==NULL) $info=$info[$what];
		return $info;
	}
	/**
	 * Gets the file group.
	 * @param string $path
	 * @return string
	 */
	final function filegroup ($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("HybridFS: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		return $this->getinfo($path,'group');
	}
	/**
	 * This function returns the time when the data blocks of a file were being written to,
	 * that is, the time when the content of the file was changed.
	 * @param string $path
	 * @return int UNIX timestamp
	 */
	final function filemtime($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("HybridFS: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		return $this->getinfo($path,'time');
	}
	/**
	 * Gets the file owner.
	 * @param string $path
	 * @return string
	 */
	final function fileowner($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("HybridFS: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		return $this->getinfo($path,'owner');
	}
	/**
	 * Gets permissions for the given file.
	 * @param string $path
	 * @return string The octal rights on the file
	 */
	final function fileperms($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("HybridFS: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		return $this->getinfo($path,'rights');
	}
	/**
	 * Gets the size for the given file in bytes
	 * @param string $path
	 * @return int the filesize in bytes
	 */
	final function filesize($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("HybridFS: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		return $this->getinfo($path,'size');
	}
	/**
	 * Returns the type of the given file.
	 * @param string $path
	 * @return string 'file' if the path is a file, 'directory' if the path is a directory
	 */
	final function filetype($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("HybridFS: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		return ($this->getinfo($path,'type')=="f"?'file':'directory');
	}
	/**
	 * Checks wether a path is a file or not.
	 * @param string $path
	 * @return bool true if the path is a file, false if not.
	 */
	final function is_file($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("HybridFS: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->file_exists($path)) return false;
		return ($this->getinfo($path,'type')=="f");
	}
	/**
	 * Checks wether a path if a directory or not.
	 * @param string $path
	 * @return bool true if the path is a directory, false if not.
	 */
	final function is_dir($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("HybridFS: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->file_exists($path)) return false;
		return ($this->getinfo($path,'type')=="d");
	}
	/**
	 * Returns contents of the file at $path or false if $path isn't a file
	 * @param string $path
	 * @return mixed The file's contents if it is a file, false if it isn't a file.
	 */
	final function file_get_contents($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("HybridFS: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->is_readable($path)) { $this->trigger_error("HybridFS: $path is not readable",E_USER_WARNING); return false; }
		if(!$this->is_file($path)) { $this->trigger_error("HybridFS: $path is not a file",E_USER_WARNING); return false; }
		$filename=$this->basename($path);
		$dir=$this->dirname($path);
		if($dir=='.'||$dir=="\\") $dir="";
		$result=$this->connection->query("SELECT * FROM `{$this->table}` WHERE `path`='$dir' AND `filename` = '$filename'")->fetch_assoc();
		return $this->bin64K_getFile($result['mfi']);
	}
	/**
	 * Reads an entire file into an array.
	 * @param string $path
	 * @param int $flags
	 * @return array
	 */
	final function file($path,$flags=0) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("HybridFS: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->is_readable($path)) { $this->trigger_error("HybridFS: $path is not readable",E_USER_WARNING); return false; }
		if(!$this->is_file($path)) { $this->trigger_error("HybridFS: $path is not a file",E_USER_WARNING); return false; }
		$c=$this->file_get_contents($path);
		$c=str_replace("\r\n","\n",$c);
		$c=str_replace("\r","\n",$c);
		if($flags==(4||6)) {
			$c=preg_replace("/\n+/","\n",$c);
			$c=preg_replace("/\n$/","",$c);
		}
		$arr=explode("\n",$c);
		if($flags!=2&&$flags!=6) {
			foreach($arr as &$v) $v.="\n";
		}
		return $arr;
		
	}
	/**
	 * Checks if a directory is empty
	 * @param string $path
	 * @return bool true if $path is empty, false if the directory is not empty
	 * @internal returns true when a file is given
	 */
	final function is_empty($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("HybridFS: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->is_readable($path)) { $this->trigger_error("HybridFS: $path is not readable",E_USER_WARNING); return false; }
		if(!$this->is_dir($path)) return true; //File doesn't contain files
		$dir=$this->dirname($path);
		$result=$this->connection->query("SELECT * FROM `{$this->table}` WHERE `path`='{$path}'");
		$r=$result->num_rows;
		$result->free();
		return ($r==0);
	}
	/**
	 * Removes a directory
	 * @param string $path
	 * @return bool Returns true on success and false on failure
	 */
	final function rmdir($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("HybridFS: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->is_writable($path)) { $this->trigger_error("HybridFS: $path is not writable",E_USER_WARNING); return false; }
		if(!$this->is_dir($path)) { $this->trigger_error("HybridFS: '$path' is not a directory",E_USER_WARNING); return false; }
		if(!$this->is_empty($path)) { $this->trigger_error("HybridFS: '$path' is not empty",E_USER_WARNING); return false; }
		$filename=$this->basename($path);
		$dir=$this->dirname($path);
		if($dir=="."||$dir=="\\") $dir='';
		return $this->connection->query("DELETE FROM `{$this->table}` WHERE `path`='$dir' AND `filename` = '$filename'");
	}
	/**
	 * Removes a file
	 * @param string $path
	 * @return bool true on success, false on failure
	 */
	final function unlink($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("HybridFS: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->is_writable($path)) { $this->trigger_error("HybridFS: $path is not writable",E_USER_WARNING); return false; }
		if(!$this->is_file($path)) { $this->trigger_error("HybridFS: '$path' is not a file",E_USER_WARNING); return false; }
		$filename=$this->basename($path);
		$dir=$this->dirname($path);
		if($dir=="."||$dir=="\\") $dir='';
		$this->bin64K_delete($this->getInfo($path,'mfi'));
		return $this->connection->query("DELETE FROM `{$this->table}` WHERE `path`='$dir' AND `filename` = '$filename'");
	}
	/**
	 * Checks wether a file exists
	 * @param string $path
	 * @return bool true if $path exists, false if $path doesn't exist
	 */
	final function file_exists($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("HybridFS: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		$filename=$this->basename($path);
		$dir=$this->dirname($path);
		if($dir=="."||$dir=="\\") $dir='';
		$result=$this->connection->query("SELECT * FROM `{$this->table}` WHERE `path`='$dir' AND `filename` = '$filename'");
		$r=$result->num_rows;
		$result->free();
		return ($r==1);
	}
	/**
	 * Creates a directory
	 * @param string $path
	 * @param int $mode Octal permissions of the file. If not defined, it takes the permissions of the parent directory
	 * @return bool
	 */
	final function mkdir($path,$mode=NULL) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("HybridFS: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if($this->file_exists($path)) { $this->trigger_error("HybridFS: $path already exists",E_USER_WARNING); return false; } //File exists
		$filename=$this->basename($path);
		$dir=$this->dirname($path);
		if(!$this->is_writable($dir)) { $this->trigger_error("HybridFS: $dir is not writable",E_USER_WARNING); return false; }
		if(!$this->is_dir($dir)) { $this->trigger_error("HybridFS: '$dir' does not exist",E_USER_WARNING); return false; } //Dir above doesn't exist
		if($dir=="."||$dir=="\\") $dir='';
		if($mode===NULL) $mode=$this->getinfo($dir,'rights'); //Get mode from .. dir
		return $this->connection->query("INSERT INTO `{$this->table}` VALUES ('$dir', '$filename', '$mode', '{$this->user}', '{$this->group}', CURRENT_TIMESTAMP, 'd', 'application/octet-stream',0,'')");
	}
	/**
	 * Creates a file
	 * @param string $path
	 * @param int $mode Octal permissions of the file. If not defined, it takes the permissions of the parent directory
	 * @param string $mime The mimetype of the file, if not defined, the mimetype will be looked up by extention
	 * @return bool
	 */
	 final function mkfile($path,$mode=NULL,$mime=NULL) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("HybridFS: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if($this->file_exists($path)) { $this->trigger_error("HybridFS: $path already exists",E_USER_WARNING); return false; } //File exists
		$filename=$this->basename($path);
		$dir=$this->dirname($path);
		if(!$this->is_writable($dir)) { $this->trigger_error("HybridFS: $dir is not writable",E_USER_WARNING); return false; }
		if($dir=="."||$dir=="\\") $dir='';
		if(!$this->is_dir($dir)) { $this->trigger_error("HybridFS: '$dir' does not exist",E_USER_WARNING); return false; } //Dir above doesn't exist
		if($mode===NULL) $mode=$this->getinfo($dir,'rights'); // Get mode from .. dir
		if($mime===NULL) $mime=$this->mime_content_type($path); // Get mime from extention
		return $this->connection->query("INSERT INTO `{$this->table}` VALUES ('$dir', '$filename', '$mode', '{$this->user}', '{$this->group}', CURRENT_TIMESTAMP, 'f', '$mime',0,'')");
	}
	/**
	 * Writes data to the file
	 * @param string $path The path to write to
	 * @param string $data The data to write to the file
	 * @return bool
	 */
	final function file_put_contents($path,$data) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("HybridFS: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->is_writable($path)) { $this->trigger_error("HybridFS: $path is not writable",E_USER_WARNING); return false; }
		if(!$this->file_exists($path)) $this->mkfile($path);
		if(!$this->is_file($path)) { $this->trigger_error("HybridFS: '$path' is not a file",E_USER_WARNING); return false; }
		$filename=$this->basename($path);
		$dir=$this->dirname($path);
		if($dir=="."||$dir=="\\") $dir='';
		return ($this->bin64K_uploadString($data,$this->getInfo($path,'mfi'))!=0);
	}
	/**
	 * Checks wether a file is readable
	 * @param string $path
	 * @return bool true is the file is readable, false otherwise
	 */
	final function is_readable($path){
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("HybridFS: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->file_exists($path)) return false;
		$rights=$this->getinfo($path,'rights');
		return (($this->getinfo($path,'owner')==$this->user&&$rights[0]>=4)
				||(($this->getinfo($path,'group')==$this->group)
				  ||in_array($this->getinfo($path,'group'),$this->groups)
				  &&$rights[1]>=4)
				||($rights[2]>=4));
	}
	/**
	 * Checks if the file is writeable
	 * @param string $path
	 * @return bool true if the file is writeable, false otherwise
	 */
	final function is_writable($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("HybridFS: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->file_exists($path)) return false;
		$rights=$this->getinfo($path,'rights');
		for($i=0;$i<=2;$i++) {
			if($rights[$i]>=4) $rights[$i]=$rights[$i]-4;
		}
		return (($this->getinfo($path,'owner')==$this->user&&$rights[0]>=2)
				||(($this->getinfo($path,'group')==$this->group)
				  ||in_array($this->getinfo($path,'group'),$this->groups)
				  &&$rights[1]>=2)
				||($rights[2]>=2));
	}
	/**
	 * Checks if the file is executable by php
	 * @param string $path
	 * @return bool true if the file is executable, false otherwise
	 */
	final function is_executable($path){
		$this->trigger_error('HybridFS: is_executable is deprecated', E_USER_DEPRECATED);
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("HybridFS: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->file_exists($path)) return false;
		$rights=$this->getinfo($path,'rights');
		return (($this->getinfo($path,'owner')==$this->user&&$rights[0]%2==1)
				||(($this->getinfo($path,'group')==$this->group)
				  ||in_array($this->getinfo($path,'group'),$this->groups)
				  &&$rights[1]%2==1)
				||($rights[2]%2==1));
	}
	/**
	 * You can add some extra security here
	 * To write (once) supply array with key-value pairs
	 * To read, supply the array key
	 * @param string|array $key
	 * @return mixed
	 * @deprecated 2.0
	 */
	final function security($key) {
		$this->trigger_error('HybridFS: security is deprecated', E_USER_DEPRECATED);
	}
	/**
	 * Change permissions
	 * @param string $path
	 * @param int $mode The new permissions for the file
	 * @return bool
	 */
	final function chmod($path,$mode) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("HybridFS: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->is_writable($path)) { $this->trigger_error("HybridFS: File cannot be written",E_USER_WARNING); return false; }
		$filename=$this->basename($path);
		$dir=$this->dirname($path);
		if($dir=='.'||$dir=='\\') $dir='';
		return $this->connection->query("UPDATE `{$this->table}` SET `rights`='{$this->connection->escape_string($mode)}' WHERE `path` = '$dir' AND `filename` = '$filename'");
	}
	/**
	 * Change owner
	 * @param string $path
	 * @param string $user New user of the file
	 * @uses HybridFS_validate::valid_user_function()
	 * @return bool
	 */
	final function chown($path,$user) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("HybridFS: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if($this->is_writable($path)) { $this->trigger_error("HybridFS: $path is not writable",E_USER_WARNING); return false; }
		if(!$this->valid_user_function($user)) { $this->trigger_error("HybridFS: User does not exist",E_USER_WARNING); return false; }
		$filename=$this->basename($path);
		$dir=$this->dirname($path);
		if($dir=='.'||$dir=='\\') $dir='';
		return $this->connection->query("UPDATE `{$this->table}` SET `owner`='{$this->connection->escape_string($user)}' WHERE `path` = '$dir' AND `filename` = '$filename'");
	}
	/**
	 * Changes group
	 * @param string $path
	 * @param string $group New group of the file
	 * @uses HybridFS_validate::valid_group_function()
	 * @return bool
	 */
	final function chgrp($path,$group) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("HybridFS: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->is_writable($path)) { $this->trigger_error("HybridFS: $path is not writable",E_USER_WARNING); return false; }
		if(!$this->valid_group_function($group)) { $this->trigger_error("HybridFS: Group does not exist",E_USER_WARNING); return false; }
		$filename=$this->basename($path);
		$dir=$this->dirname($path);
		if($dir=='.'||$dir=='\\') $dir='';
		return $this->connection->query("UPDATE `{$this->table}` SET `group`='{$this->connection->escape_string($group)}' WHERE `path` = '$dir' AND `filename` = '$filename'");
	}
	/**
	 * Rename a file
	 * @param string $oldpath The current location of the file
	 * @param string $newpath The new location of the file
	 * @return bool
	 * @internal You must have reading and writing permissions on $oldpath and writing permissions on $newpath
	 */
	final function rename($oldpath,$newpath) {
		if(($oldpath=$this->simplify_path($oldpath))===NULL)  { $this->trigger_error("HybridFS: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(($newpath=$this->simplify_path($newpath))===NULL)  { $this->trigger_error("HybridFS: ".func_get_arg(1)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->is_readable($oldpath)) { $this->trigger_error("HybridFS: $oldpath is not readable",E_USER_WARNING); return false; }
		if(!$this->is_writable($oldpath)) { $this->trigger_error("HybridFS: $olpath is not writable", E_USER_WARNING); return false;	}
		$oldfilename=$this->basename($oldpath);
		$newfilename=$this->basename($newpath);
		$olddir=$this->dirname($oldpath);
		$newdir=$this->dirname($newpath);
		if(!$this->is_dir($newdir)) $newdir=$this->simplify_path($olddir.'/'.$newdir); //Relative?
		if(!$this->is_writable($newdir)) { $this->trigger_error("HybridFS: $newdir is not writable",E_USER_WARNING); return false; }
		if($this->file_exists($newpath)) { $this->trigger_error("HybridFS: $newpath already exists",E_USER_WARNING); return false; }
		if(preg_match("|^$oldpath|",$newpath)) { $this->trigger_error("HybridFS: $newpath is a subdirectory of $oldpath",E_USER_WARNING); return false; }
		if($olddir=='.'||$olddir=='\\') $olddir='';
		if($newdir=='.'||$newdir=='\\') $newdir='';
		$result=$this->connection->query("SELECT * FROM `{$this->table}` WHERE `path` LIKE '{$this->simplify_path($olddir.'/'.$oldfilename)}/%' OR `path`='{$this->simplify_path($olddir.'/'.$oldfilename)}'");
		$x=array();
		while($row=$result->fetch_assoc()) {
			$row['newpath']=str_replace($oldpath,$newpath,$row['path']);
			$x[]=$this->connection->query("UPDATE `{$this->table}` SET `path`='$row[newpath]' WHERE `path` = '$row[path]'");
		}
		$result->free();
		$x[]=$this->connection->query("UPDATE `{$this->table}` SET `path`='$newdir',`filename`='$newfilename' WHERE `path` = '$olddir' AND `filename` = '$oldfilename'");
		return !in_array(false,$x);
	}
	/**
	 * Copy a file to another location
	 * @param string $source The current location of the file
	 * @param string $dest The location you want the file to be copied to
	 * @internal You must have reading permissions on $source and writing permissions on $dest
	 * @return bool
	 */
	final function copy($source,$dest) {
		if(($source=$this->simplify_path($source))===NULL)  { $this->trigger_error("HybridFS: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(($dest=$this->simplify_path($dest))===NULL)  { $this->trigger_error("HybridFS: ".func_get_arg(1)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->is_readable($source)) { $this->trigger_error("HybridFS: $source is not readable",E_USER_WARNING); return false; }
		$sfilename=$this->basename($source);
		$dfilename=$this->basename($dest);
		$sdir=$this->dirname($source);
		$ddir=$this->dirname($dest);
		if(!$this->is_writable($ddir)) { $this->trigger_error("HybridFS: $ddir is not writable",E_USER_WARNING); return false; }
		if($this->file_exists($dest)) { $this->trigger_error("HybridFS: $dest already exists",E_USER_WARNING); return false; }
		if($sdir=='.'||$sdir=='\\') $sdir='';
		if($ddir=='.'||$ddir=='\\') $ddir='';
		$x=array();
		$result=$this->connection->query("SELECT * FROM `{$this->table}` WHERE `path` LIKE '$source/%' OR `path` = '$source'");
		while($row=$result->fetch_assoc()) {
			$row['path']=str_replace($source,$dest,$row['path']);
			$x[]=$this->connection->query("INSERT into `{$this->table}` VALUES ('{$row['path']}', '{$row['filename']}','{$row['rights']}','{$row['owner']}','{$row['group']}','{$row['time']}','{$row['type']}','{$row['mime']}','{$row['size']}',NULL)");
			$this->bin64K_uploadString($this->bin64K_getFile($row['mfi']),$this->connection->insert_id);
		}
		$result->free();
		$result=$this->connection->query("SELECT * FROM `{$this->table}` WHERE `path` = '$sdir' AND `filename`='$sfilename'");
		$row=$result->fetch_assoc();
		$result->free();
		$row['path']=str_replace($sdir,$ddir,$row['path']);
		$row['filename']=str_replace($sfilename,$dfilename,$row['filename']);
		$x[]=$this->connection->query("INSERT into `{$this->table}` VALUES ('{$row['path']}', '{$row['filename']}','{$row['rights']}','{$row['owner']}','{$row['group']}','{$row['time']}','{$row['type']}','{$row['mime']}','{$row['size']}',NULL)");
		$this->bin64K_uploadString($this->bin64K_getFile($row['mfi']),$this->connection->insert_id);
		return !in_array(false,$x);
	}
	/**
	 * Moves a newly uploaded file to its final location
	 * @param string $source The source path on the real filesystem
	 * @param string $dest The destiniation on HybridFS
	 * @internal This function checks to ensure that the file designated by filename is a valid upload file (meaning that it was uploaded via PHP's HTTP POST upload mechanism).
	 * @internal If the file is valid, it will be moved to the filename given by destination.
	 * @return bool
	 */
	final function move_uploaded_file($source,$dest) {
		if(!is_uploaded_file($source)) { $this->trigger_error("HybridFS: $source is not a valid upload file",E_USER_WARNING); return false; }
		if(($dest=$this->simplify_path($dest))===NULL)  { $this->trigger_error("HybridFS: ".func_get_arg(1)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->file_exists($dest)) $this->mkfile($dest);
		return $this->file_put_contents($dest,file_get_contents($source));
	}
	/**
	 * Reads a file and writes it to the output buffer
	 * @param string $path
	 * @return int The length of the written string
	 */
	final function readfile($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("HybridFS: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		$c=$this->file_get_contents($path);
		echo $c;
		return strlen($c);
	}
	/**
	 * Returns an array of files and directories from the directory.
	 * @param string $path
	 * @param int $sorting_order sort ascending if omitted or 0, descending if 1
	 * @return array Array containing the filenames
	 */
	final function scandir($path,$sorting_order=0) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("HybridFS: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->file_exists($path)) { $this->trigger_error("HybridFS: $path does not exist",E_USER_WARNING); return false; }
		if(!$this->is_dir($path)) { $this->trigger_error("HybridFS: $path is not a directory",E_USER_WARNING); return false; }
		if(!$this->is_readable($path)) { $this->trigger_error("HybridFS: $path is not readable",E_USER_WARNING); return false; }
		$res=$this->connection->query("SELECT * FROM `{$this->table}` WHERE `path`='$path' ORDER BY `filename` ".(($sorting_order==0)?"ASC":"DESC"));
		$files=array();
		$files[]='.';
		$files[]='..';
		while($f=$res->fetch_array()) {
			$files=$f['filename'];
		}
		return $files;
	}
	/**
	 * Method to insert data into the data folder
	 * @param string $data The raw data to upload
	 * @param int $id The id of the fileindex referencing to this data
	 * @return int Id of the fileindex
	 */
	private final function bin64K_uploadString($data,$id){
		if(!$this->realfs->file_put_contents($id, $data)) $this->trigger_error('HybridFS: Could not write data to file', E_USER_WARNING);
		return $id;
	}
	/**
	 * Method to fetch the data from the data folder
	 * @param int $id The id of the fileindex referencing to this data
	 * @return string The whole file
	 */
	private final function bin64K_getFile($id) {
		if(!$this->realfs->file_exists($id)) return '';
		return $this->realfs->file_get_contents($id);
	}

	/**
	 * Deletes chunk entries in the data table
	 * @param int $id The id of the fileindex
	 * @return bool
	 */
	private final function bin64K_delete($id) {
		return $this->realfs->unlink($id);
	}
	/**
	 * Prevents errors when no validating class is provided
	 * @param mixed $user
	 * @return bool false (always fails)
	 */
	function valid_user_function($user) {
		return false;
	}
	/**
	 * Prevents errors when no validating class is provided
	 * @param mixed $group
	 * @return bool false (always fails)
	 */
	function valid_group_function($group) {
		return false;
	}
}

/**
 * This interface is used to construct user and group checking classes
 * @author Lars Vierbergen
 * @package HybridFS
 */
interface HybridFS_validate {
	/**
	 * This function checks wether a user is valid
	 * @param string $user The user to check
	 * @return bool true if the user is valid, false if not
	 */
	function valid_user_function($user);
	/**
	 * This function checks wether a group is valid
	 * @param string $group The group to check
	 * @return bool true if the group is valid, false if not
	 */
	function valid_group_function($group);
}
?>