<?php //@./inc/mysqlfs/mysqlfs.php ?>
<?php
/**
 * The main library file for MySQLfs
 * @package mysqlfs
 */
/**
 * This is the interface to MySQlfs
 * @author Lars Vierbergen
 * @package mysqlfs
 */
class mysqlfs {
	/**
	 * The main group of the user
	 * @var string
	 */
	public $group=0;
	/**
	 * The userid if the users
	 * @var string
	 */
	public $user=NULL;
	/**
	 * All groups the user is a member of
	 * @var array
	 */
	public $groups=array();
	/**
	 * Holds the security settings
	 * @var array
	 * @deprecated 2.0
	 */
	protected $security_settings=array();
	/**
	 * Holds the MySQL connection
	 * @var mysqli
	 */
	protected $connection;
	/**
	 * Holds the table for MySQLfs
	 * @var string
	 */
	protected $table;
	/**
	 * Startup function for the class. Registers MySQL connection and table settings
	 * @param MySQLi $conn The MySQLi link to the database
	 * @param string $tbl The name of the index table, the datatable will have 'data' appended
	 */
	function __construct($conn,$tbl='index') {
		$this->connection=$conn;
		$this->table=$tbl;
		/**
		 * @internal Avoid warnings about strtotime calls (bug #815161)
		 */
		date_default_timezone_set(@date_default_timezone_get());
	}
	/**
	 * Sets $user and $group (user's main group) and $groups (all groups user is member of)
	 * @param string $user The identifier of the current user
	 * @param string $group The identifier of the main group of the user
	 * @param array $groups All groups the user is member of
	 */
	final function setuser($user=NULL,$group=0,$groups=array()) {
		$this->user=$user;
		$this->group=$group;
		$this->groups=$groups;
	}
	/**
	 * Reads the directory into an array
	 * @param string $path The path to read
	 * @return array Array with nested arrays containing information about the file
	 */
	final function readdir_full($path){
		if(!$this->is_dir($path)) { $this->trigger_error("MySQLfs: $path is not a directory",E_USER_WARNING); return NULL; }
		if(!$this->is_readable($path)) { $this->trigger_error("MySQLfs: $path is not readable",E_USER_WARNING); return NULL; }
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		$result=$this->connection->query("SELECT `type`,`filename` , `time` , `rights`, `size` , `owner` , `group` FROM `{$this->table}` WHERE `path` = '$path' AND `filename`!=''");
		$out=array();
		$out[]=$this->getinfo($path.'/..'); $out[0]['filename']='..';
		$out[]=$this->getinfo($path); $out[1]['filename']='.';
		while($row=$result->fetch_assoc()) {
			if(!$this->is_readable($path.'/'.$row['filename'])) continue;
			$out[]=$row;
		}
		$result->free();
		return $out;
	}
	/**
	 * Converts bytes to human-readable file sizes
	 * @param int $size The size in bytes to convert
	 * @return string
	 */
	final function sizetohuman($size) {
		$sizes = array(" Bytes", " KB", " MB", " GB", " TB", " PB", " EB", " ZB", " YB");
		if ($size == 0) { return('n/a'); } else {
		return (round($size/pow(1024, ($i = floor(log($size, 1024)))), 2) . $sizes[$i]);
		}
	}
	/**
	 * Gets information about a path
	 * @param string $path The path where to retrieve the information from
	 * @param string $what Which field to return
	 * @return mixed array when $what is not defined, string when $what is defined
	 */
	final function getinfo($path,$what=NULL) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		$filename=$this->basename($path);
		$dir=$this->dirname($path);
		if($dir=="\\"||$dir=='.') $dir="";
		$result=$this->connection->query("SELECT `type` , `filename` , `time` , `rights` , `mime` , `size`, `mfi`, `owner` , `group` FROM `{$this->table}` WHERE `path`='$dir' AND `filename` = '$filename'");
		$info=$result->fetch_assoc();
		$result->free();
		if($info['owner']===NULL) $info['owner']='nobody';
		if($info['group']===NULL) $info['group']='nobody';
		if($info['type']=="d") unset($info['mime']); //Dir's have no mimetype
		if($what!==NULL) $info=$info[$what];
		return $info;
	}
	/**
	 * Gets the file group.
	 * @param string $path
	 * @return string
	 */
	final function filegroup ($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		return $this->getinfo($path,'group');
	}
	/**
	 * This function returns the time when the data blocks of a file were being written to,
	 * that is, the time when the content of the file was changed.
	 * @param string $path
	 * @return int UNIX timestamp
	 */
	final function filemtime($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		return $this->getinfo($path,'time');
	}
	/**
	 * Gets the file owner.
	 * @param string $path
	 * @return string
	 */
	final function fileowner($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		return $this->getinfo($path,'owner');
	}
	/**
	 * Gets permissions for the given file.
	 * @param string $path
	 * @return string The octal rights on the file
	 */
	final function fileperms($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		return $this->getinfo($path,'rights');
	}
	/**
	 * Gets the size for the given file in bytes
	 * @param string $path
	 * @return int the filesize in bytes
	 */
	final function filesize($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		return $this->getinfo($path,'size');
	}
	/**
	 * Returns the type of the given file.
	 * @param string $path
	 * @return string 'file' if the path is a file, 'directory' if the path is a directory
	 */
	final function filetype($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		return ($this->getinfo($path,'type')=="f"?'file':'directory');
	}
	/**
	 * Removes /./ and resolves /../
	 * @param string $path
	 * @return string The new path
	 * @internal Returns NULL when an invalid path is given
	 */
	final function simplify_path($path) {
		$path='/'.$path.'/';
		if(!preg_match('@^([^\0])+$@',$path)) {  return NULL; }
		$parts = array_filter(explode('/', $path), 'strlen');
		$absolutes = array();
		foreach ($parts as $part) {
			if ('.' == $part) continue;
			if ('..' == $part) {
				array_pop($absolutes);
			} else {
				$absolutes[] = $part;
			}
		}
		if(($ret=implode('/', $absolutes))=='')  {
			return '/';
		}
		else {
			return $ret;
		}
	}
	/**
	 * Checks wether a path is a file or not.
	 * @param string $path
	 * @return bool true if the path is a file, false if not.
	 */
	final function is_file($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->file_exists($path)) return false;
		return ($this->getinfo($path,'type')=="f");
	}
	/**
	 * Checks wether a path if a directory or not.
	 * @param string $path
	 * @return bool true if the path is a directory, false if not.
	 */
	final function is_dir($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->file_exists($path)) return false;
		return ($this->getinfo($path,'type')=="d");
	}
	/**
	 * Returns contents of the file at $path or false if $path isn't a file
	 * @param string $path
	 * @return mixed The file's contents if it is a file, false if it isn't a file.
	 */
	final function file_get_contents($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->is_readable($path)) { $this->trigger_error("MySQLfs: $path is not readable",E_USER_WARNING); return false; }
		if(!$this->is_file($path)) { $this->trigger_error("MySQLfs: $path is not a file",E_USER_WARNING); return false; }
		$path=$this->connection->escape_string($path);
		$filename=$this->basename($path);
		$dir=$this->dirname($path);
		if($dir=='.'||$dir=="\\") $dir="";
		$result=$this->connection->query("SELECT `mfi` FROM `{$this->table}` WHERE `path`='$dir' AND `filename` = '$filename'")->fetch_assoc();
		return $this->bin64K_getFile($result['mfi']);
	}
	/**
	 * Reads an entire file into an array.
	 * @param string $path
	 * @param int $flags
	 * @return array
	 */
	final function file($path,$flags=0) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->is_readable($path)) { $this->trigger_error("MySQLfs: $path is not readable",E_USER_WARNING); return false; }
		if(!$this->is_file($path)) { $this->trigger_error("MySQLfs: $path is not a file",E_USER_WARNING); return false; }
		$c=$this->file_get_contents($path);
		$c=str_replace("\r\n","\n",$c);
		$c=str_replace("\r","\n",$c);
		if($flags==(4||6)) {
			$c=preg_replace("/\n+/","\n",$c);
			$c=preg_replace("/\n$/","",$c);
		}
		$arr=explode("\n",$c);
		if($flags!=2&&$flags!=6) {
			foreach($arr as &$v) $v.="\n";
		}
		return $arr;
		
	}
	/**
	 * Checks if a directory is empty
	 * @param string $path
	 * @return bool true if $path is empty, false if the directory is not empty
	 * @internal returns true when a file is given
	 */
	final function is_empty($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->is_readable($path)) { $this->trigger_error("MySQLfs: $path is not readable",E_USER_WARNING); return false; }
		if(!$this->is_dir($path)) return true; //File doesn't contain files
		$dir=$this->dirname($path);
		$result=$this->connection->query("SELECT `rights` FROM `{$this->table}` WHERE `path`='{$path}'");
		$r=$result->num_rows;
		$result->free();
		return ($r==0);
	}
	/**
	 * Removes a directory
	 * @param string $path
	 * @return bool Returns true on success and false on failure
	 */
	final function rmdir($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->is_writable($path)) { $this->trigger_error("MySQLfs: $path is not writable",E_USER_WARNING); return false; }
		if(!$this->is_dir($path)) { $this->trigger_error("MySQLfs: '$path' is not a directory",E_USER_WARNING); return false; }
		if(!$this->is_empty($path)) { $this->trigger_error("MySQLfs: '$path' is not empty",E_USER_WARNING); return false; }
		$filename=$this->basename($path);
		$dir=$this->dirname($path);
		if($dir=="."||$dir=="\\") $dir='';
		return $this->connection->query("DELETE FROM `{$this->table}` WHERE `path`='$dir' AND `filename` = '$filename'");
	}
	/**
	 * Removes a file
	 * @param string $path
	 * @return bool true on success, false on failure
	 */
	final function unlink($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->is_writable($path)) { $this->trigger_error("MySQLfs: $path is not writable",E_USER_WARNING); return false; }
		if(!$this->is_file($path)) { $this->trigger_error("MySQLfs: '$path' is not a file",E_USER_WARNING); return false; }
		$filename=$this->basename($path);
		$dir=$this->dirname($path);
		if($dir=="."||$dir=="\\") $dir='';
		$this->bin64K_delete($this->getInfo($path,'mfi'));
		return $this->connection->query("DELETE FROM `{$this->table}` WHERE `path`='$dir' AND `filename` = '$filename'");
	}
	/**
	 * Checks wether a file exists
	 * @param string $path
	 * @return bool true if $path exists, false if $path doesn't exist
	 */
	final function file_exists($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		$filename=$this->basename($path);
		$dir=$this->dirname($path);
		if($dir=="."||$dir=="\\") $dir='';
		$result=$this->connection->query("SELECT `rights` FROM `{$this->table}` WHERE `path`='$dir' AND `filename` = '$filename'");
		$r=$result->num_rows;
		$result->free();
		return ($r==1);
	}
	/**
	 * Creates a directory
	 * @param string $path
	 * @param int $mode Octal permissions of the file. If not defined, it takes the permissions of the parent directory
	 * @return bool
	 */
	final function mkdir($path,$mode=NULL) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if($this->file_exists($path)) { $this->trigger_error("MySQLfs: $path already exists",E_USER_WARNING); return false; } //File exists
		$filename=$this->basename($path);
		$dir=$this->dirname($path);
		if(!$this->is_writable($dir)) { $this->trigger_error("MySQLfs: $dir is not writable",E_USER_WARNING); return false; }
		if(!$this->is_dir($dir)) { $this->trigger_error("MySQLfs: '$dir' does not exist",E_USER_WARNING); return false; } //Dir above doesn't exist
		if($dir=="."||$dir=="\\") $dir='';
		if($mode===NULL) $mode=$this->getinfo($dir,'rights'); //Get mode from .. dir
		return $this->connection->query("INSERT INTO `{$this->table}` VALUES ('$dir', '$filename', '$mode', '{$this->user}', '{$this->group}', CURRENT_TIMESTAMP, 'd', 'application/octet-stream',0,'')");
	}
	/**
	 * Creates a file
	 * @param string $path
	 * @param int $mode Octal permissions of the file. If not defined, it takes the permissions of the parent directory
	 * @param string $mime The mimetype of the file, if not defined, the mimetype will be looked up by extention
	 * @return bool
	 */
	 final function mkfile($path,$mode=NULL,$mime=NULL) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if($this->file_exists($path)) { $this->trigger_error("MySQLfs: $path already exists",E_USER_WARNING); return false; } //File exists
		$filename=$this->basename($path);
		$dir=$this->dirname($path);
		if(!$this->is_writable($dir)) { $this->trigger_error("MySQLfs: $dir is not writable",E_USER_WARNING); return false; }
		if($dir=="."||$dir=="\\") $dir='';
		if(!$this->is_dir($dir)) { $this->trigger_error("MySQLfs: '$dir' does not exist",E_USER_WARNING); return false; } //Dir above doesn't exist
		if($mode===NULL) $mode=$this->getinfo($dir,'rights'); // Get mode from .. dir
		if($mime===NULL) $mime=$this->mime_content_type($path); // Get mime from extention
		return $this->connection->query("INSERT INTO `{$this->table}` VALUES ('$dir', '$filename', '$mode', '{$this->user}', '{$this->group}', CURRENT_TIMESTAMP, 'f', '$mime',0,'')");
	}
	/**
	 * Gets the content type based on the extention
	 * @param string $path
	 * @return string mimetype of the file
	 */
	final function mime_content_type($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		$mime_types = array("323" => "text/h323","acx" => "application/internet-property-stream","ai" => "application/postscript","aif" => "audio/x-aiff","aifc" => "audio/x-aiff","aiff" => "audio/x-aiff","asf" => "video/x-ms-asf","asr" => "video/x-ms-asf","asx" => "video/x-ms-asf","au" => "audio/basic","avi" => "video/x-msvideo","axs" => "application/olescript","bas" => "text/plain","bcpio" => "application/x-bcpio","bin" => "application/octet-stream","bmp" => "image/bmp","c" => "text/plain","cat" => "application/vnd.ms-pkiseccat","cdf" => "application/x-cdf","cer" => "application/x-x509-ca-cert","class" => "application/octet-stream","clp" => "application/x-msclip","cmx" => "image/x-cmx","cod" => "image/cis-cod","cpio" => "application/x-cpio","crd" => "application/x-mscardfile","crl" => "application/pkix-crl","crt" => "application/x-x509-ca-cert","csh" => "application/x-csh","css" => "text/css","dcr" => "application/x-director","der" => "application/x-x509-ca-cert","dir" => "application/x-director","dll" => "application/x-msdownload","dms" => "application/octet-stream","doc" => "application/msword","dot" => "application/msword","dvi" => "application/x-dvi","dxr" => "application/x-director","eps" => "application/postscript","etx" => "text/x-setext","evy" => "application/envoy","exe" => "application/octet-stream","fif" => "application/fractals","flr" => "x-world/x-vrml","gif" => "image/gif","gtar" => "application/x-gtar","gz" => "application/x-gzip","h" => "text/plain","hdf" => "application/x-hdf","hlp" => "application/winhlp","hqx" => "application/mac-binhex40","hta" => "application/hta","htc" => "text/x-component","htm" => "text/html","html" => "text/html","htt" => "text/webviewhtml","ico" => "image/x-icon","ief" => "image/ief","iii" => "application/x-iphone","ins" => "application/x-internet-signup","isp" => "application/x-internet-signup","jfif" => "image/pipeg","jpe" => "image/jpeg","jpeg" => "image/jpeg","jpg" => "image/jpeg","js" => "application/x-javascript","latex" => "application/x-latex","lha" => "application/octet-stream","lsf" => "video/x-la-asf","lsx" => "video/x-la-asf","lzh" => "application/octet-stream","m13" => "application/x-msmediaview","m14" => "application/x-msmediaview","m3u" => "audio/x-mpegurl","man" => "application/x-troff-man","mdb" => "application/x-msaccess","me" => "application/x-troff-me","mht" => "message/rfc822","mhtml" => "message/rfc822","mid" => "audio/mid","mny" => "application/x-msmoney","mov" => "video/quicktime","movie" => "video/x-sgi-movie","mp2" => "video/mpeg","mp3" => "audio/mpeg","mpa" => "video/mpeg","mpe" => "video/mpeg","mpeg" => "video/mpeg","mpg" => "video/mpeg","mpp" => "application/vnd.ms-project","mpv2" => "video/mpeg","ms" => "application/x-troff-ms","mvb" => "application/x-msmediaview","nws" => "message/rfc822","oda" => "application/oda","p10" => "application/pkcs10","p12" => "application/x-pkcs12","p7b" => "application/x-pkcs7-certificates","p7c" => "application/x-pkcs7-mime","p7m" => "application/x-pkcs7-mime","p7r" => "application/x-pkcs7-certreqresp","p7s" => "application/x-pkcs7-signature","pbm" => "image/x-portable-bitmap","pdf" => "application/pdf","pfx" => "application/x-pkcs12","pgm" => "image/x-portable-graymap","pko" => "application/ynd.ms-pkipko","pma" => "application/x-perfmon","pmc" => "application/x-perfmon","pml" => "application/x-perfmon","pmr" => "application/x-perfmon","pmw" => "application/x-perfmon","pnm" => "image/x-portable-anymap","pot" => "application/vnd.ms-powerpoint","ppm" => "image/x-portable-pixmap","pps" => "application/vnd.ms-powerpoint","ppt" => "application/vnd.ms-powerpoint","prf" => "application/pics-rules","ps" => "application/postscript","pub" => "application/x-mspublisher","qt" => "video/quicktime","ra" => "audio/x-pn-realaudio","ram" => "audio/x-pn-realaudio","ras" => "image/x-cmu-raster","rgb" => "image/x-rgb","rmi" => "audio/mid","roff" => "application/x-troff","rtf" => "application/rtf","rtx" => "text/richtext","scd" => "application/x-msschedule","sct" => "text/scriptlet","setpay" => "application/set-payment-initiation","setreg" => "application/set-registration-initiation","sh" => "application/x-sh","shar" => "application/x-shar","sit" => "application/x-stuffit","snd" => "audio/basic","spc" => "application/x-pkcs7-certificates","spl" => "application/futuresplash","src" => "application/x-wais-source","sst" => "application/vnd.ms-pkicertstore","stl" => "application/vnd.ms-pkistl","stm" => "text/html","svg" => "image/svg+xml","sv4cpio" => "application/x-sv4cpio","sv4crc" => "application/x-sv4crc","t" => "application/x-troff","tar" => "application/x-tar","tcl" => "application/x-tcl","tex" => "application/x-tex","texi" => "application/x-texinfo","texinfo" => "application/x-texinfo","tgz" => "application/x-compressed","tif" => "image/tiff","tiff" => "image/tiff","tr" => "application/x-troff","trm" => "application/x-msterminal","tsv" => "text/tab-separated-values","txt" => "text/plain","uls" => "text/iuls","ustar" => "application/x-ustar","vcf" => "text/x-vcard","vrml" => "x-world/x-vrml","wav" => "audio/x-wav","wcm" => "application/vnd.ms-works","wdb" => "application/vnd.ms-works","wks" => "application/vnd.ms-works","wmf" => "application/x-msmetafile","wps" => "application/vnd.ms-works","wri" => "application/x-mswrite","wrl" => "x-world/x-vrml","wrz" => "x-world/x-vrml","xaf" => "x-world/x-vrml","xbm" => "image/x-xbitmap","xla" => "application/vnd.ms-excel","xlc" => "application/vnd.ms-excel","xlm" => "application/vnd.ms-excel","xls" => "application/vnd.ms-excel","xlt" => "application/vnd.ms-excel","xlw" => "application/vnd.ms-excel","xof" => "x-world/x-vrml","xpm" => "image/x-xpixmap","xwd" => "image/x-xwindowdump","z" => "application/x-compress","zip" => "application/zip");
		$filename=$this->basename($path);
		$parts=explode('.',$filename);
		$ext=$parts[count($parts)-1];
		if(!array_key_exists($ext,$mime_types))  return 'application/octet-stream'; //Not in array? Must be default!
		return $mime_types[$ext];
	}
	/**
	 * Writes data to the file
	 * @param string $path The path to write to
	 * @param string $data The data to write to the file
	 * @return bool
	 */
	final function file_put_contents($path,$data) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->is_writable($path)) { $this->trigger_error("MySQLfs: $path is not writable",E_USER_WARNING); return false; }
		if(!$this->file_exists($path)) $this->mkfile($path);
		if(!$this->is_file($path)) { $this->trigger_error("MySQLfs: '$path' is not a file",E_USER_WARNING); return false; }
		$filename=$this->basename($path);
		$dir=$this->dirname($path);
		if($dir=="."||$dir=="\\") $dir='';
		return ($this->bin64K_uploadString($data,$this->getInfo($path,'mfi'))!=0);
	}
	/**
	 * Checks wether a file is readable
	 * @param string $path
	 * @return bool true is the file is readable, false otherwise
	 */
	final function is_readable($path){
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->file_exists($path)) return false;
		$rights=$this->getinfo($path,'rights');
		return (($this->getinfo($path,'owner')==$this->user&&$rights[0]>=4)
				||(($this->getinfo($path,'group')==$this->group)
				  ||in_array($this->getinfo($path,'group'),$this->groups)
				  &&$rights[1]>=4)
				||($rights[2]>=4));
	}
	/**
	 * Checks if the file is writeable
	 * @param string $path
	 * @return bool true if the file is writeable, false otherwise
	 */
	final function is_writable($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->file_exists($path)) return false;
		$rights=$this->getinfo($path,'rights');
		for($i=0;$i<=2;$i++) {
			if($rights[$i]>=4) $rights[$i]=$rights[$i]-4;
		}
		return (($this->getinfo($path,'owner')==$this->user&&$rights[0]>=2)
				||(($this->getinfo($path,'group')==$this->group)
				  ||in_array($this->getinfo($path,'group'),$this->groups)
				  &&$rights[1]>=2)
				||($rights[2]>=2));
	}
	/**
	 * Alias of is_writable
	 * @param string $path
	 * @return bool
	 * @ignore
	 * @see mysqlfs::is_writeable()
	 */
	final function is_writeable($path) {
		return $this->is_writable($path);
	}
	/**
	 * Checks if the file is executable by php
	 * @param string $path
	 * @return bool true if the file is executable, false otherwise
	 * @deprecated 2.0
	 */
	final function is_executable($path){
		$this->trigger_error('MySQLfs: is_executable is deprecated', E_USER_DEPRECATED);
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->file_exists($path)) return false;
		$rights=$this->getinfo($path,'rights');
		return (($this->getinfo($path,'owner')==$this->user&&$rights[0]%2==1)
				||(($this->getinfo($path,'group')==$this->group)
				  ||in_array($this->getinfo($path,'group'),$this->groups)
				  &&$rights[1]%2==1)
				||($rights[2]%2==1));
	}
	/**
	 * You can add some extra security here
	 * To write (once) supply array with key-value pairs
	 * To read, supply the array key
	 * @param string|array $key
	 * @return mixed
	 * @deprecated 2.0
	 */
	final function security($key) {
		$this->trigger_error('MySQLfs: security is deprecated', E_USER_DEPRECATED);
	}
	/**
	 * Change permissions
	 * @param string $path
	 * @param int $mode The new permissions for the file
	 * @return bool
	 */
	final function chmod($path,$mode) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->is_writable($path)) { $this->trigger_error("MySQLfs: File cannot be written",E_USER_WARNING); return false; }
		$filename=$this->basename($path);
		$dir=$this->dirname($path);
		if($dir=='.'||$dir=='\\') $dir='';
		return $this->connection->query("UPDATE `{$this->table}` SET `rights`='{$this->connection->escape_string($mode)}' WHERE `path` = '$dir' AND `filename` = '$filename'");
	}
	/**
	 * Change owner
	 * @param string $path
	 * @param string $user New user of the file
	 * @uses mysqlfs_validate::valid_user_function()
	 * @return bool
	 */
	final function chown($path,$user) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if($this->is_writable($path)) { $this->trigger_error("MySQLfs: $path is not writable",E_USER_WARNING); return false; }
		if(!$this->valid_user_function($user)) { $this->trigger_error("MySQLfs: User does not exist",E_USER_WARNING); return false; }
		$filename=$this->basename($path);
		$dir=$this->dirname($path);
		if($dir=='.'||$dir=='\\') $dir='';
		return $this->connection->query("UPDATE `{$this->table}` SET `owner`='{$this->connection->escape_string($user)}' WHERE `path` = '$dir' AND `filename` = '$filename'");
	}
	/**
	 * Changes group
	 * @param string $path
	 * @param string $group New group of the file
	 * @uses mysqlfs_validate::valid_group_function()
	 * @return bool
	 */
	final function chgrp($path,$group) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->is_writable($path)) { $this->trigger_error("MySQLfs: $path is not writable",E_USER_WARNING); return false; }
		if(!$this->valid_group_function($group)) { $this->trigger_error("MySQLfs: Group does not exist",E_USER_WARNING); return false; }
		$filename=$this->basename($path);
		$dir=$this->dirname($path);
		if($dir=='.'||$dir=='\\') $dir='';
		return $this->connection->query("UPDATE `{$this->table}` SET `group`='{$this->connection->escape_string($group)}' WHERE `path` = '$dir' AND `filename` = '$filename'");
	}
	/**
	 * Rename a file
	 * @param string $oldpath The current location of the file
	 * @param string $newpath The new location of the file
	 * @return bool
	 * @internal You must have reading and writing permissions on $oldpath and writing permissions on $newpath
	 */
	final function rename($oldpath,$newpath) {
		if(($oldpath=$this->simplify_path($oldpath))===NULL)  { $this->trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(($newpath=$this->simplify_path($newpath))===NULL)  { $this->trigger_error("MySQLfs: ".func_get_arg(1)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->is_readable($oldpath)) { $this->trigger_error("MySQLfs: $oldpath is not readable",E_USER_WARNING); return false; }
		if(!$this->is_writable($oldpath)) { $this->trigger_error("MySQLfs: $olpath is not writable", E_USER_WARNING); return false;	}
		$oldfilename=$this->basename($oldpath);
		$newfilename=$this->basename($newpath);
		$olddir=$this->dirname($oldpath);
		$newdir=$this->dirname($newpath);
		if(!$this->is_dir($newdir)) $newdir=$this->simplify_path($olddir.'/'.$newdir); //Relative?
		if(!$this->is_writable($newdir)) { $this->trigger_error("MySQLfs: $newdir is not writable",E_USER_WARNING); return false; }
		if($this->file_exists($newpath)) { $this->trigger_error("MySQLfs: $newpath already exists",E_USER_WARNING); return false; }
		if(preg_match("|^$oldpath|",$newpath)) { $this->trigger_error("MySQLfs: $newpath is a subdirectory of $oldpath",E_USER_WARNING); return false; }
		if($olddir=='.'||$olddir=='\\') $olddir='';
		if($newdir=='.'||$newdir=='\\') $newdir='';
		$result=$this->connection->query("SELECT `path` FROM `{$this->table}` WHERE `path` LIKE '{$this->simplify_path($olddir.'/'.$oldfilename)}/%' OR `path`='{$this->simplify_path($olddir.'/'.$oldfilename)}'");
		$x=array();
		while($row=$result->fetch_assoc()) {
			$row['newpath']=str_replace($oldpath,$newpath,$row['path']);
			$x[]=$this->connection->query("UPDATE `{$this->table}` SET `path`='$row[newpath]' WHERE `path` = '$row[path]'");
		}
		$result->free();
		$x[]=$this->connection->query("UPDATE `{$this->table}` SET `path`='$newdir',`filename`='$newfilename' WHERE `path` = '$olddir' AND `filename` = '$oldfilename'");
		return !in_array(false,$x);
	}
	/**
	 * Copy a file to another location
	 * @param string $source The current location of the file
	 * @param string $dest The location you want the file to be copied to
	 * @internal You must have reading permissions on $source and writing permissions on $dest
	 * @return bool
	 */
	final function copy($source,$dest) {
		if(($source=$this->simplify_path($source))===NULL)  { $this->trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(($dest=$this->simplify_path($dest))===NULL)  { $this->trigger_error("MySQLfs: ".func_get_arg(1)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->is_readable($source)) { $this->trigger_error("MySQLfs: $source is not readable",E_USER_WARNING); return false; }
		$sfilename=$this->basename($source);
		$dfilename=$this->basename($dest);
		$sdir=$this->dirname($source);
		$ddir=$this->dirname($dest);
		if(!$this->is_writable($ddir)) { $this->trigger_error("MySQLfs: $ddir is not writable",E_USER_WARNING); return false; }
		if($this->file_exists($dest)) { $this->trigger_error("MySQLfs: $dest already exists",E_USER_WARNING); return false; }
		if($sdir=='.'||$sdir=='\\') $sdir='';
		if($ddir=='.'||$ddir=='\\') $ddir='';
		$x=array();
		$result=$this->connection->query("SELECT * FROM `{$this->table}` WHERE `path` LIKE '$source/%' OR `path` = '$source'");
		while($row=$result->fetch_assoc()) {
			$row['path']=str_replace($source,$dest,$row['path']);
			$x[]=$this->connection->query("INSERT into `{$this->table}` VALUES ('{$row['path']}', '{$row['filename']}','{$row['rights']}','{$row['owner']}','{$row['group']}','{$row['time']}','{$row['type']}','{$row['mime']}','{$row['size']}',NULL)");
			$this->bin64K_uploadString($this->bin64K_getFile($row['mfi']),$this->connection->insert_id);
		}
		$result->free();
		$result=$this->connection->query("SELECT * FROM `{$this->table}` WHERE `path` = '$sdir' AND `filename`='$sfilename'");
		$row=$result->fetch_assoc();
		$result->free();
		$row['path']=str_replace($sdir,$ddir,$row['path']);
		$row['filename']=str_replace($sfilename,$dfilename,$row['filename']);
		$x[]=$this->connection->query("INSERT into `{$this->table}` VALUES ('{$row['path']}', '{$row['filename']}','{$row['rights']}','{$row['owner']}','{$row['group']}','{$row['time']}','{$row['type']}','{$row['mime']}','{$row['size']}',NULL)");
		$this->bin64K_uploadString($this->bin64K_getFile($row['mfi']),$this->connection->insert_id);
		return !in_array(false,$x);
	}
	/**
	 * Moves a newly uploaded file to its final location
	 * @param string $source The source path on the real filesystem
	 * @param string $dest The destiniation on MySQLfs
	 * @internal This function checks to ensure that the file designated by filename is a valid upload file (meaning that it was uploaded via PHP's HTTP POST upload mechanism).
	 * @internal If the file is valid, it will be moved to the filename given by destination.
	 * @return bool
	 */
	final function move_uploaded_file($source,$dest) {
		if(!is_uploaded_file($source)) { $this->trigger_error("MySQLfs: $source is not a valid upload file",E_USER_WARNING); return false; }
		if(($dest=$this->simplify_path($dest))===NULL)  { $this->trigger_error("MySQLfs: ".func_get_arg(1)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->file_exists($dest)) $this->mkfile($dest);
		return $this->file_put_contents($dest,file_get_contents($source));
	}
	/**
	 * Reads a file and writes it to the output buffer
	 * @param string $path
	 * @return int The length of the written string
	 */
	final function readfile($path) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		$c=$this->file_get_contents($path);
		echo $c;
		return strlen($c);
	}
	/**
	 * Method to insert data into the data table,
	 * creating 'chunk' entries in the data table, linked to the fileindex
	 * @param string $data The raw data to upload
	 * @param int $id The id of the fileindex referencing to this data
	 * @author Paul Scott
	 * @return int Id of the fileindex
	 */
	private final function bin64K_uploadString($data,$id){
		//Update filesize
		$this->connection->query("UPDATE `{$this->table}` SET `size`=".strlen($data)." WHERE `mfi`=$id");
		//Remove old files
		if (!$this->connection->query("DELETE FROM `{$this->table}data` WHERE `masterid`=$id")) {
			$this->trigger_error("MySQLfs: Failure to delete binary inode data!",E_USER_WARNING);
			return 0;
		}
		$data=base64_encode($data); //Base64 encode file
		// Insert into the filedata table
		while (strlen($data)>0) {
			// Make the data mysql insert safe
			$binarydata = substr($data,0,65535);
			if (!$this->connection->query("INSERT INTO `{$this->table}data` values(NULL,$id,'$binarydata')"))
			{
				$this->trigger_error("MySQLfs: Failure to insert binary inode data row!",E_USER_WARNING);
				return 0;
			}
			$data=substr($data,65535);
		}
		return $id;
	}
	/**
	 * Get the file chunchs and paste it together
	 * @param int $id The id of the fileindex referencing to this data
	 * @author Paul Scott
	 * @return string The whole file
	 */
	private final function bin64K_getFile($id) {
		// Loop through and stream the nodes 1 by 1
		$data='';
		$res=$this->connection->query("SELECT `data` FROM `{$this->table}data` WHERE `masterid` = $id ORDER BY `id`");
		while($block=$res->fetch_assoc()) {
			$data.=$block['data'];
		}
		return base64_decode($data);
	}
	/**
	 * Returns an array of files and directories from the directory.
	 * @param string $path
	 * @param int $sorting_order sort ascending if omitted or 0, descending if 1
	 * @return array Array containing the filenames
	 */
	final function scandir($path,$sorting_order=0) {
		if(($path=$this->simplify_path($path))===NULL)  { $this->trigger_error("MySQLfs: ".func_get_arg(0)." is an invalid pathname",E_USER_WARNING); return false; }
		if(!$this->file_exists($path)) { $this->trigger_error("MySQLfs: $path does not exist",E_USER_WARNING); return false; }
		if(!$this->is_dir($path)) { $this->trigger_error("MySQLfs: $path is not a directory",E_USER_WARNING); return false; }
		if(!$this->is_readable($path)) { $this->trigger_error("MySQLfs: $path is not readable",E_USER_WARNING); return false; }
		$res=$this->connection->query("SELECT `filename` FROM `{$this->table}` WHERE `path`='$path' ORDER BY `filename` ASC");
		$files=array();
		$files[]='.';
		$files[]='..';
		while($f=$res->fetch_array()) {
			$files=$f['filename'];
		}
		if($sorting_order!=0) return array_reverse($files);
		return $files;
	}
	/**
	 * Deletes chunk entries in the data table
	 * @param int $id The id of the fileindex
	 * @return bool
	 */
	private final function bin64K_delete($id) {
		return $this->connection->query("DELETE FROM `{$this->table}data` WHERE `masterid` = $id");
	}
	/**
	 * Raise a php error
	 * @param string $error The errorstring
	 * @param int $level The errorlevel
	 * @internal Remove the trigger_error line to make MySQLfs silent
	 */
	private function trigger_error($error,$level) {
		trigger_error($error,$level);
	}
	/**
	 * Given a string containing the path of a file or directory, this function will return the parent directory's path.
	 * @param string $path A path
	 * @return string The path with trailing /component removed
	 * @internal Just executes some dumb regexps
	 */
	final function dirname($path) {
		$ret=dirname($path);
		$ret=str_replace('\\', '/', $ret);
		if($ret=='.') $ret='/';
		return $ret;
	}
	/**
	 * Given a string containing the path to a file or directory, this function will return the trailing name component.
	 * @param string $path A path
	 * @param string $suffix If the name components ends in $suffix, this will also be cut off.
	 * @return string The basename of the given path
	 * @internal Just executes some dumb regexps
	 */
	final function basename($path,$suffix=NULL) {
		return basename($path,$suffix);
	}
	/**
	}
	/**
	 * Prevents errors when no validating class is provided
	 * @param mixed $user
	 * @return bool false (always fails)
	 */
	function valid_user_function($user) {
		return false;
	}
	/**
	 * Prevents errors when no validating class is provided
	 * @param mixed $group
	 * @return bool false (always fails)
	 */
	function valid_group_function($group) {
		return false;
	}
}

/**
 * This interface is used to construct user and group checking classes
 * @author Lars Vierbergen
 * @package mysqlfs
 */
interface mysqlfs_validate {
	/**
	 * This function checks wether a user is valid
	 * @param string $user The user to check
	 * @return bool true if the user is valid, false if not
	 */
	function valid_user_function($user);
	/**
	 * This function checks wether a group is valid
	 * @param string $group The group to check
	 * @return bool true if the group is valid, false if not
	 */
	function valid_group_function($group);
}
?>